@@ -15,19 +15,21 @@
  */
 
 
-#include "mongoc/mongoc.h"
-#include "mongoc/mongoc-apm-private.h"
-#include "mongoc/mongoc-counters-private.h"
-#include "mongoc/mongoc-client-pool-private.h"
-#include "mongoc/mongoc-client-pool.h"
-#include "mongoc/mongoc-client-private.h"
-#include "mongoc/mongoc-queue-private.h"
-#include "mongoc/mongoc-thread-private.h"
-#include "mongoc/mongoc-topology-private.h"
-#include "mongoc/mongoc-trace-private.h"
+#include "mongoc.h"
+#include "mongoc-apm-private.h"
+#include "mongoc-counters-private.h"
+#include "mongoc-client-pool-private.h"
+#include "mongoc-client-pool.h"
+#include "mongoc-client-private.h"
+#include "mongoc-client-side-encryption-private.h"
+#include "mongoc-queue-private.h"
+#include "mongoc-thread-private.h"
+#include "mongoc-topology-private.h"
+#include "mongoc-topology-background-monitoring-private.h"
+#include "mongoc-trace-private.h"
 
 #ifdef MONGOC_ENABLE_SSL
-#include "mongoc/mongoc-ssl-private.h"
+#include "mongoc-ssl-private.h"
 #endif
 
 struct _mongoc_client_pool_t {
@@ -56,17 +58,16 @@ void
 mongoc_client_pool_set_ssl_opts (mongoc_client_pool_t *pool,
                                  const mongoc_ssl_opt_t *opts)
 {
-   BSON_ASSERT (pool);
-
    bson_mutex_lock (&pool->mutex);
 
-   _mongoc_ssl_opts_cleanup (&pool->ssl_opts);
+   _mongoc_ssl_opts_cleanup (&pool->ssl_opts,
+                             false /* don't free internal opts. */);
 
-   memset (&pool->ssl_opts, 0, sizeof pool->ssl_opts);
    pool->ssl_opts_set = false;
 
    if (opts) {
-      _mongoc_ssl_opts_copy_to (opts, &pool->ssl_opts);
+      _mongoc_ssl_opts_copy_to (
+         opts, &pool->ssl_opts, false /* don't overwrite internal opts. */);
       pool->ssl_opts_set = true;
    }
 
@@ -75,6 +76,20 @@ mongoc_client_pool_set_ssl_opts (mongoc_client_pool_t *pool,
 
    bson_mutex_unlock (&pool->mutex);
 }
+
+void
+_mongoc_client_pool_set_internal_tls_opts (
+   mongoc_client_pool_t *pool, _mongoc_internal_tls_opts_t *internal)
+{
+   bson_mutex_lock (&pool->mutex);
+   if (!pool->ssl_opts_set) {
+      return;
+   }
+   pool->ssl_opts.internal = bson_malloc (sizeof (_mongoc_internal_tls_opts_t));
+   memcpy (
+      pool->ssl_opts.internal, internal, sizeof (_mongoc_internal_tls_opts_t));
+   bson_mutex_unlock (&pool->mutex);
+}
 #endif
 
 
@@ -93,7 +108,7 @@ mongoc_client_pool_new (const mongoc_uri_t *uri)
    BSON_ASSERT (uri);
 
 #ifndef MONGOC_ENABLE_SSL
-   if (mongoc_uri_get_ssl (uri)) {
+   if (mongoc_uri_get_tls (uri)) {
       MONGOC_ERROR ("Can't create SSL client pool,"
                     " SSL not enabled in this build.");
       return NULL;
@@ -139,12 +154,14 @@ mongoc_client_pool_new (const mongoc_uri_t *uri)
    }
 
 #ifdef MONGOC_ENABLE_SSL
-   if (mongoc_uri_get_ssl (pool->uri)) {
+   if (mongoc_uri_get_tls (pool->uri)) {
       mongoc_ssl_opt_t ssl_opt = {0};
+      _mongoc_internal_tls_opts_t internal_tls_opts = {0};
 
-      _mongoc_ssl_opts_from_uri (&ssl_opt, pool->uri);
+      _mongoc_ssl_opts_from_uri (&ssl_opt, &internal_tls_opts, pool->uri);
       /* sets use_ssl = true */
       mongoc_client_pool_set_ssl_opts (pool, &ssl_opt);
+      _mongoc_client_pool_set_internal_tls_opts (pool, &internal_tls_opts);
    }
 #endif
    mongoc_counter_client_pools_active_inc ();
@@ -182,7 +199,7 @@ mongoc_client_pool_destroy (mongoc_client_pool_t *pool)
    mongoc_cond_destroy (&pool->cond);
 
 #ifdef MONGOC_ENABLE_SSL
-   _mongoc_ssl_opts_cleanup (&pool->ssl_opts);
+   _mongoc_ssl_opts_cleanup (&pool->ssl_opts, true);
 #endif
 
    bson_free (pool);
@@ -202,9 +219,10 @@ mongoc_client_pool_destroy (mongoc_client_pool_t *pool)
 static void
 _start_scanner_if_needed (mongoc_client_pool_t *pool)
 {
-   if (!_mongoc_topology_start_background_scanner (pool->topology)) {
-      MONGOC_ERROR ("Background scanner did not start!");
-      abort ();
+   if (!pool->topology->single_threaded) {
+      bson_mutex_lock (&pool->topology->mutex);
+      _mongoc_topology_background_monitoring_start (pool->topology);
+      bson_mutex_unlock (&pool->topology->mutex);
    }
 }
 
@@ -232,11 +250,21 @@ mongoc_client_t *
 mongoc_client_pool_pop (mongoc_client_pool_t *pool)
 {
    mongoc_client_t *client;
+   int32_t wait_queue_timeout_ms;
+   int64_t expire_at_ms = -1;
+   int64_t now_ms;
+   int r;
 
    ENTRY;
 
    BSON_ASSERT (pool);
 
+   wait_queue_timeout_ms = mongoc_uri_get_option_as_int32 (
+      pool->uri, MONGOC_URI_WAITQUEUETIMEOUTMS, -1);
+   if (wait_queue_timeout_ms > 0) {
+      expire_at_ms =
+         (bson_get_monotonic_time () / 1000) + wait_queue_timeout_ms;
+   }
    bson_mutex_lock (&pool->mutex);
 
 again:
@@ -246,12 +274,26 @@ again:
          _initialize_new_client (pool, client);
          pool->size++;
       } else {
-         mongoc_cond_wait (&pool->cond, &pool->mutex);
+         if (wait_queue_timeout_ms > 0) {
+            now_ms = bson_get_monotonic_time () / 1000;
+            if (now_ms < expire_at_ms) {
+               r = mongoc_cond_timedwait (
+                  &pool->cond, &pool->mutex, expire_at_ms - now_ms);
+               if (mongo_cond_ret_is_timedout (r)) {
+                  GOTO (done);
+               }
+            } else {
+               GOTO (done);
+            }
+         } else {
+            mongoc_cond_wait (&pool->cond, &pool->mutex);
+         }
          GOTO (again);
       }
    }
 
    _start_scanner_if_needed (pool);
+done:
    bson_mutex_unlock (&pool->mutex);
 
    RETURN (client);
@@ -453,3 +495,12 @@ mongoc_client_pool_set_appname (mongoc_client_pool_t *pool, const char *appname)
 
    return ret;
 }
+
+bool
+mongoc_client_pool_enable_auto_encryption (mongoc_client_pool_t *pool,
+                                           mongoc_auto_encryption_opts_t *opts,
+                                           bson_error_t *error)
+{
+   return _mongoc_cse_client_pool_enable_auto_encryption (
+      pool->topology, opts, error);
+}
